from aerosandbox.dynamics.point_mass.common_point_mass import _DynamicsPointMassBaseClass
from aerosandbox.weights.mass_properties import MassProperties
import aerosandbox.numpy as np
from typing import Union, Dict, Tuple


class DynamicsPointMass3DSpeedGammaTrack(_DynamicsPointMassBaseClass):
    """
    Dynamics instance:
    * simulating a point mass
    * in 3D
    * with velocity parameterized in speed-gamma-track space

    State variables:
        x_e: x-position, in Earth axes. [meters]
        y_e: y-position, in Earth axes. [meters]
        z_e: z-position, in Earth axes. [meters]
        speed: Speed; equivalent to u_w, the x-velocity in wind axes. [m/s]
        gamma: Flight path angle. [radians]
        track: Track angle. [radians]
            * Track of 0 == North == aligned with x_e axis
            * Track of np.pi / 2 == East == aligned with y_e axis

    Indirect control variables:
        alpha: Angle of attack. [degrees]
        beta: Sideslip angle. [degrees]
        bank: Bank angle. [radians]

    Control variables:
        Fx_w: Force along the wind-x axis. [N]
        Fy_w: Force along the wind-y axis. [N]
        Fz_w: Force along the wind-z axis. [N]

    """

    def __init__(self,
                 mass_props: MassProperties = None,
                 x_e: Union[np.ndarray, float] = 0,
                 y_e: Union[np.ndarray, float] = 0,
                 z_e: Union[np.ndarray, float] = 0,
                 speed: Union[np.ndarray, float] = 0,
                 gamma: Union[np.ndarray, float] = 0,
                 track: Union[np.ndarray, float] = 0,
                 alpha: Union[np.ndarray, float] = 0,
                 beta: Union[np.ndarray, float] = 0,
                 bank: Union[np.ndarray, float] = 0,
                 ):
        # Initialize state variables
        self.mass_props = MassProperties() if mass_props is None else mass_props
        self.x_e = x_e
        self.y_e = y_e
        self.z_e = z_e
        self.speed = speed
        self.gamma = gamma
        self.track = track

        # Initialize indirect control variables
        self.alpha = alpha
        self.beta = beta
        self.bank = bank

        # Initialize control variables
        self.Fx_w = 0
        self.Fy_w = 0
        self.Fz_w = 0

    @property
    def state(self) -> Dict[str, Union[float, np.ndarray]]:
        return {
            "x_e"  : self.x_e,
            "y_e"  : self.y_e,
            "z_e"  : self.z_e,
            "speed": self.speed,
            "gamma": self.gamma,
            "track": self.track,
        }

    @property
    def control_variables(self) -> Dict[str, Union[float, np.ndarray]]:
        return {
            "alpha": self.alpha,
            "beta" : self.beta,
            "bank" : self.bank,
            "Fx_w" : self.Fx_w,
            "Fy_w" : self.Fy_w,
            "Fz_w" : self.Fz_w,
        }

    def state_derivatives(self) -> Dict[str, Union[float, np.ndarray]]:

        d_speed = self.Fx_w / self.mass_props.mass

        sb = np.sin(self.bank)
        cb = np.cos(self.bank)

        force_gamma_direction = -cb * self.Fz_w - sb * self.Fy_w  # Force in the direction that acts to increase gamma
        force_track_direction = -sb * self.Fz_w + cb * self.Fy_w  # Force in the direction that acts to increase track

        d_gamma = force_gamma_direction / self.mass_props.mass / self.speed
        d_track = force_track_direction / self.mass_props.mass / self.speed / np.cos(self.gamma)

        return {
            "x_e"  : self.u_e,
            "y_e"  : self.v_e,
            "z_e"  : self.w_e,
            "speed": d_speed,
            "gamma": d_gamma,
            "track": d_track,
        }

    @property
    def u_e(self):
        return self.speed * np.cos(self.gamma) * np.cos(self.track)

    @property
    def v_e(self):
        return self.speed * np.cos(self.gamma) * np.sin(self.track)

    @property
    def w_e(self):
        return -self.speed * np.sin(self.gamma)

    def convert_axes(self,
                     x_from: float,
                     y_from: float,
                     z_from: float,
                     from_axes: str,
                     to_axes: str,
                     ) -> Tuple[float, float, float]:
        if (from_axes == "earth" or to_axes == "earth"):
            rot_w_to_e = np.rotation_matrix_from_euler_angles(
                roll_angle=self.bank,
                pitch_angle=self.gamma,
                yaw_angle=self.track,
                as_array=False
            )

        if from_axes == "wind":
            x_w = x_from
            y_w = y_from
            z_w = z_from
        elif from_axes == "earth":
            x_w = rot_w_to_e[0][0] * x_from + rot_w_to_e[1][0] * y_from + rot_w_to_e[2][0] * z_from
            y_w = rot_w_to_e[0][1] * x_from + rot_w_to_e[1][1] * y_from + rot_w_to_e[2][1] * z_from
            z_w = rot_w_to_e[0][2] * x_from + rot_w_to_e[1][2] * y_from + rot_w_to_e[2][2] * z_from
        else:
            x_w, y_w, z_w = self.op_point.convert_axes(
                x_from, y_from, z_from,
                from_axes=from_axes, to_axes="wind"
            )

        if to_axes == "wind":
            x_to = x_w
            y_to = y_w
            z_to = z_w
        elif to_axes == "earth":
            x_to = rot_w_to_e[0][0] * x_w + rot_w_to_e[0][1] * y_w + rot_w_to_e[0][2] * z_w
            y_to = rot_w_to_e[1][0] * x_w + rot_w_to_e[1][1] * y_w + rot_w_to_e[1][2] * z_w
            z_to = rot_w_to_e[2][0] * x_w + rot_w_to_e[2][1] * y_w + rot_w_to_e[2][2] * z_w
        else:
            x_to, y_to, z_to = self.op_point.convert_axes(
                x_w, y_w, z_w,
                from_axes="wind", to_axes=to_axes
            )

        return x_to, y_to, z_to

    def add_force(self,
                  Fx: Union[np.ndarray, float] = 0,
                  Fy: Union[np.ndarray, float] = 0,
                  Fz: Union[np.ndarray, float] = 0,
                  axes="wind",
                  ) -> None:
        Fx_w, Fy_w, Fz_w = self.convert_axes(
            x_from=Fx,
            y_from=Fy,
            z_from=Fz,
            from_axes=axes,
            to_axes="wind"
        )
        self.Fx_w = self.Fx_w + Fx_w
        self.Fy_w = self.Fy_w + Fy_w
        self.Fz_w = self.Fz_w + Fz_w


if __name__ == '__main__':
    dyn = DynamicsPointMass3DSpeedGammaTrack()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

u_e_0 = 100
v_e_0 = 0
w_e_0 = -100
speed_0 = (u_e_0 ** 2 + w_e_0 ** 2) ** 0.5
gamma_0 = np.arctan2(-w_e_0, u_e_0)
track_0 = 0

time = np.linspace(0, 10, 501)
N = len(time)


def get_trajectory(
        parameterization: type = asb.DynamicsPointMass3DCartesian,
        gravity=True,
        drag=True,
        sideforce=True,
        plot=False,
        verbose=False,
):
    opti = asb.Opti()

    if parameterization is asb.DynamicsPointMass3DCartesian:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=opti.variable(np.linspace(0, 300, N)),
            y_e=opti.variable(np.linspace(0, 0, N), scale=1),
            z_e=opti.variable(np.linspace(0, 0, N), scale=100),
            u_e=opti.variable(np.linspace(100, 50, N)),
            v_e=opti.variable(np.linspace(0, 0, N), scale=1),
            w_e=opti.variable(np.linspace(-100, 50, N)),
        )
    elif parameterization is asb.DynamicsPointMass3DSpeedGammaTrack:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=opti.variable(np.linspace(0, 300, N)),
            y_e=opti.variable(np.linspace(0, 0, N), scale=1),
            z_e=opti.variable(np.linspace(0, 0, N), scale=100),
            speed=opti.variable(np.linspace(100, 50, N)),
            gamma=opti.variable(np.linspace(0, 0, N)),
            track=opti.variable(np.linspace(0, 0, N)),
        )
    else:
        raise ValueError("Bad value of `parameterization`!")

    if gravity:
        dyn.add_gravity_force()
    q = 0.5 * 1.225 * dyn.speed ** 2
    if drag:
        dyn.add_force(
            Fx=-1 * (0.1) ** 2 * q,
            axes="wind"
        )
    if sideforce:
        strouhal = 0.2
        frequency = 5  # strouhal * dyn.speed / 0.1
        dyn.add_force(
            Fy=q * 1 * (0.1) ** 2 * np.sin(2 * np.pi * frequency * time),
            axes="wind"
        )

    dyn.constrain_derivatives(
        opti=opti,
        time=time,
    )

    opti.subject_to([
        dyn.x_e[0] == 0,
        dyn.y_e[0] == 0,
        dyn.z_e[0] == 0,
        dyn.u_e[0] == 100,
        dyn.v_e[0] == 0,
        dyn.w_e[0] == -100,
    ])

    sol = opti.solve(verbose=verbose)

    dyn.substitute_solution(sol)

    if plot:
        import matplotlib.pyplot as plt;
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots()
        p.plot_color_by_value(dyn.x_e, dyn.altitude, c=dyn.speed, colorbar=True)
        p.equal()
        p.show_plot("Trajectory", "$x_e$", "$z_e$")

    return dyn


def test_final_position_Cartesian_with_sideforce():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True,
        sideforce=True
    )
    assert dyn[-1].x_e == pytest.approx(277.3463197415092, abs=1e-1)
    assert dyn[-1].y_e == pytest.approx(10.791223276048788, abs=1)
    assert dyn[-1].z_e == pytest.approx(3.1722726974414805, abs=1e-1)


def test_final_position_Cartesian_with_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(277.5774436945314, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722727033631886, abs=1e-2)


def test_final_position_Cartesian_no_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=False,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(1000, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(-509.5, abs=1e-2)


def test_final_position_SpeedGammaTrack_with_sideforce():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True,
        sideforce=True
    )
    assert dyn[-1].x_e == pytest.approx(277.3463197415092, abs=1e-1)
    assert dyn[-1].y_e == pytest.approx(10.791223276048788, abs=1)
    assert dyn[-1].z_e == pytest.approx(3.1722726974414805, abs=1e-1)


def test_final_position_SpeedGammaTrack_with_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(277.5774436945314, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(3.1722727033631886, abs=1e-2)


def test_final_position_SpeedGammaTrack_no_drag():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=False,
        sideforce=False
    )
    assert dyn[-1].x_e == pytest.approx(1000, abs=1e-2)
    assert dyn[-1].z_e == pytest.approx(-509.5, abs=1e-2)


def test_cross_compare_with_drag():
    dyn1 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=True
    )
    dyn2 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=True
    )
    assert dyn1[-1].x_e == pytest.approx(dyn2[-1].x_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].y_e == pytest.approx(dyn2[-1].y_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].z_e == pytest.approx(dyn2[-1].z_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].u_e == pytest.approx(dyn2[-1].u_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].v_e == pytest.approx(dyn2[-1].v_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].w_e == pytest.approx(dyn2[-1].w_e, abs=1e-2, rel=1e-2)


def test_cross_compare_no_drag():
    dyn1 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
        drag=False
    )
    dyn2 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
        drag=False
    )
    assert dyn1[-1].x_e == pytest.approx(dyn2[-1].x_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].y_e == pytest.approx(dyn2[-1].y_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].z_e == pytest.approx(dyn2[-1].z_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].u_e == pytest.approx(dyn2[-1].u_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].v_e == pytest.approx(dyn2[-1].v_e, abs=1e-2, rel=1e-2)
    assert dyn1[-1].w_e == pytest.approx(dyn2[-1].w_e, abs=1e-2, rel=1e-2)


#
if __name__ == '__main__':
    dyn = get_trajectory(
        asb.DynamicsPointMass3DSpeedGammaTrack,
        plot=True
    )
    pytest.main()
